/*
 *  CPAchecker is a tool for configurable software verification.
 *  This file is part of CPAchecker.
 *
 *  Copyright (C) 2007-2010  Dirk Beyer
 *  All rights reserved.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *  CPAchecker web page:
 *    http://cpachecker.sosy-lab.org
 */
package org.sosy_lab.cpachecker.core;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;

import org.sosy_lab.common.Classes;
import org.sosy_lab.common.LogManager;
import org.sosy_lab.common.configuration.Configuration;
import org.sosy_lab.common.configuration.InvalidConfigurationException;
import org.sosy_lab.common.configuration.Option;
import org.sosy_lab.common.configuration.Options;
import org.sosy_lab.cpachecker.core.interfaces.CPAFactory;
import org.sosy_lab.cpachecker.core.interfaces.ConfigurableProgramAnalysis;
import org.sosy_lab.cpachecker.cpa.automaton.Automaton;
import org.sosy_lab.cpachecker.cpa.automaton.AutomatonParser;
import org.sosy_lab.cpachecker.cpa.automaton.ControlAutomatonCPA;
import org.sosy_lab.cpachecker.cpa.composite.CompositeCPA;
import org.sosy_lab.cpachecker.cpa.location.LocationCPA;
import org.sosy_lab.cpachecker.exceptions.CPAException;

import com.google.common.base.Preconditions;
import com.google.common.base.Throwables;
import com.google.common.collect.ImmutableList;

@Options
public class CPABuilder {

  private static final String CPA_OPTION_NAME = "cpa";
  private static final String CPA_CLASS_PREFIX = "org.sosy_lab.cpachecker";

  @Option(name=CPA_OPTION_NAME)
  private String cpaName = CompositeCPA.class.getCanonicalName();

  @Option(name="specification", type=Option.Type.OPTIONAL_INPUT_FILE)
  private File specificationFile = null;
  
  private final Configuration config;
  private final LogManager logger;

  public CPABuilder(Configuration pConfig, LogManager pLogger) throws InvalidConfigurationException {
    this.config = pConfig;
    this.logger = pLogger;
    config.inject(this);
  }

  public ConfigurableProgramAnalysis buildCPAs() throws InvalidConfigurationException, CPAException {
    Set<String> usedAliases = new HashSet<String>();
    
    // create automata cpas for specification given in specification file
    List<ConfigurableProgramAnalysis> cpas = null;
    if (specificationFile != null) {
      List<Automaton> automata = AutomatonParser.parseAutomatonFile(specificationFile, config, logger);
      cpas = new ArrayList<ConfigurableProgramAnalysis>(automata.size());
      
      for (Automaton automaton : automata) {
        String cpaAlias = automaton.getName();
        if (!usedAliases.add(cpaAlias)) {
          throw new InvalidConfigurationException("Name " + cpaAlias + " used twice for an automaton.");
        }
        
        CPAFactory factory = ControlAutomatonCPA.factory();
        factory.setConfiguration(Configuration.copyWithNewPrefix(config, cpaAlias));
        factory.setLogger(logger);
        factory.set(automaton, Automaton.class);
        cpas.add(factory.createInstance());
        logger.log(Level.FINE, "Loaded Automaton\"" + automaton.getName() + "\"");
      }
    }
    return buildCPAs(cpaName, CPA_OPTION_NAME, usedAliases, cpas);
  }

  private ConfigurableProgramAnalysis buildCPAs(String optionValue, String optionName, Set<String> usedAliases, List<ConfigurableProgramAnalysis> cpas) throws InvalidConfigurationException, CPAException {
    Preconditions.checkNotNull(optionValue);

    // parse option (may be of syntax "classname alias"
    String[] optionParts = optionValue.trim().split("\\s+");
    String cpaName = optionParts[0];
    String cpaAlias = getCPAAlias(optionValue, optionName, optionParts, cpaName);

    if (!usedAliases.add(cpaAlias)) {
      throw new InvalidConfigurationException("Alias " + cpaAlias + " used twice for a CPA.");
    }

    // first get instance of appropriate factory

    Class<?> cpaClass = getCPAClass(optionName, cpaName);

    CPAFactory factory = getFactoryInstance(cpaName, cpaClass);

    // now use factory to get an instance of the CPA

    factory.setConfiguration(Configuration.copyWithNewPrefix(config, cpaAlias));
    factory.setLogger(logger);

    createAndSetChildrenCPAs(cpaName, cpaAlias, factory, usedAliases, cpas);
    
    if (cpas != null && !cpas.isEmpty()) {
      throw new InvalidConfigurationException("Option specification gave specification automata, but no CompositeCPA was used");
    }

    // finally call createInstance
    ConfigurableProgramAnalysis cpa;
    try {
      cpa = factory.createInstance();
    } catch (IllegalStateException e) {
      throw new InvalidConfigurationException(e.getMessage());
    }
    return cpa;
  }

  private String getCPAAlias(String optionValue, String optionName,
      String[] optionParts, String cpaName) throws InvalidConfigurationException {

    if (optionParts.length == 1) {
      // no user-specified alias, use last part of class name
      int dotIndex = cpaName.lastIndexOf('.');
      return (dotIndex >= 0 ? cpaName.substring(dotIndex+1) : cpaName);

    } else if (optionParts.length == 2) {
      return optionParts[1];

    } else {
      throw new InvalidConfigurationException("Option " + optionName + " contains invalid CPA specification \"" + optionValue + "\"!");
    }
  }

  private Class<?> getCPAClass(String optionName, String cpaName) throws InvalidConfigurationException {
    Class<?> cpaClass;
    try {
      cpaClass = Classes.forName(cpaName, CPA_CLASS_PREFIX);
    } catch (ClassNotFoundException e) {
      throw new InvalidConfigurationException("Option " + optionName + " is set to unknown CPA " + cpaName);
    }

    if (!ConfigurableProgramAnalysis.class.isAssignableFrom(cpaClass)) {
      throw new InvalidConfigurationException(
        "Option " + optionName + " has to be set to a class implementing the ConfigurableProgramAnalysis interface!");
    }
    return cpaClass;
  }

  private CPAFactory getFactoryInstance(String cpaName, Class<?> cpaClass) throws CPAException {
    Object factoryObj;
    try {
      Method factoryMethod = cpaClass.getMethod("factory", (Class<?>[]) null);
      factoryObj = factoryMethod.invoke(null, (Object[])null);
    
    } catch (NullPointerException e) {
      throw new CPAException("The factory method of the CPA " + cpaName + " is not static!");
      
    } catch (NoSuchMethodException e) {
      throw new CPAException("Each CPA class has to offer a public static method factory with zero parameters, but " + cpaName + " does not!");
    
    } catch (IllegalArgumentException e) {
      // method is not static
      throw new CPAException("Each CPA class has to offer a public static method factory with zero parameters, but " + cpaName + " does not!");
    
    } catch (IllegalAccessException e) {
      // method is not public
      throw new CPAException("Each CPA class has to offer a public static method factory with zero parameters, but " + cpaName + " does not!");

    } catch (InvocationTargetException e) {
      Throwable cause = e.getCause();
      Throwables.propagateIfPossible(cause, CPAException.class);
      
      logger.logException(Level.FINE, cause, "CPA factory methods should never throw an exception!");
      throw new CPAException("Cannot create CPA because of unexpected exception: " + cause.getMessage());
    }

    if ((factoryObj == null) || !(factoryObj instanceof CPAFactory)) {
      throw new CPAException("The factory method of a CPA has to return an instance of CPAFactory!");
    }

    return (CPAFactory)factoryObj;
  }

  private void createAndSetChildrenCPAs(String cpaName, String cpaAlias,
      CPAFactory factory, Set<String> usedAliases, List<ConfigurableProgramAnalysis> cpas) throws InvalidConfigurationException, CPAException {
    String childOptionName = cpaAlias + ".cpa";
    String childrenOptionName = cpaAlias + ".cpas";
    String childCpaName = config.getProperty(childOptionName);
    String childrenCpaNames = config.getProperty(childrenOptionName);

    if (childrenCpaNames == null && childCpaName == null && cpaAlias.equals("CompositeCPA")
        && cpas != null && !cpas.isEmpty()) {
      // if a specification was given, but no CPAs, insert a LocationCPA
      childrenCpaNames = LocationCPA.class.getCanonicalName();
    }

    if (childCpaName != null) {
      // only one child CPA
      if (childrenCpaNames != null) {
        throw new InvalidConfigurationException("Ambiguous configuration: both "
            + childOptionName + " and " + childrenOptionName + " are specified!");
      }

      ConfigurableProgramAnalysis child = buildCPAs(childCpaName, childOptionName, usedAliases, cpas);
      try {
        factory.setChild(child);
      } catch (UnsupportedOperationException e) {
        throw new InvalidConfigurationException(cpaName + " is no wrapper CPA, but option " + childOptionName + " was specified!");
      }

    } else if (childrenCpaNames != null) {
      // several children CPAs
      ImmutableList.Builder<ConfigurableProgramAnalysis> childrenCpas = ImmutableList.builder();

      for (String currentChildCpaName : childrenCpaNames.split("\\s*,\\s*")) {
        childrenCpas.add(buildCPAs(currentChildCpaName, childrenOptionName, usedAliases, null));
      }
      if (cpas != null) {
        childrenCpas.addAll(cpas);
        cpas.clear();
      }

      try {
        factory.setChildren(childrenCpas.build());
      } catch (UnsupportedOperationException e) {
        throw new InvalidConfigurationException(cpaName + " is no wrapper CPA, but option " + childrenOptionName + " was specified!");
      }
    }
  }
}
