/*
 *  CPAchecker is a tool for configurable software verification.
 *  This file is part of CPAchecker.
 *
 *  Copyright (C) 2007-2010  Dirk Beyer
 *  All rights reserved.
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *
 *  CPAchecker web page:
 *    http://cpachecker.sosy-lab.org
 */
package org.sosy_lab.cpachecker.cpa.interval;

import java.math.BigInteger;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import org.sosy_lab.common.configuration.Configuration;
import org.sosy_lab.common.configuration.InvalidConfigurationException;
import org.sosy_lab.common.configuration.Option;
import org.sosy_lab.common.configuration.Options;
import org.sosy_lab.cpachecker.cfa.ast.IASTArraySubscriptExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTAssignment;
import org.sosy_lab.cpachecker.cfa.ast.IASTBinaryExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTCastExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTExpressionStatement;
import org.sosy_lab.cpachecker.cfa.ast.IASTFieldReference;
import org.sosy_lab.cpachecker.cfa.ast.IASTFunctionCall;
import org.sosy_lab.cpachecker.cfa.ast.IASTFunctionCallAssignmentStatement;
import org.sosy_lab.cpachecker.cfa.ast.IASTFunctionCallExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTFunctionCallStatement;
import org.sosy_lab.cpachecker.cfa.ast.IASTIdExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTInitializer;
import org.sosy_lab.cpachecker.cfa.ast.IASTInitializerExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTIntegerLiteralExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTLiteralExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTPointerTypeSpecifier;
import org.sosy_lab.cpachecker.cfa.ast.IASTRightHandSide;
import org.sosy_lab.cpachecker.cfa.ast.IASTStatement;
import org.sosy_lab.cpachecker.cfa.ast.IASTUnaryExpression;
import org.sosy_lab.cpachecker.cfa.ast.IASTBinaryExpression.BinaryOperator;
import org.sosy_lab.cpachecker.cfa.ast.IASTUnaryExpression.UnaryOperator;
import org.sosy_lab.cpachecker.cfa.objectmodel.CFAEdge;
import org.sosy_lab.cpachecker.cfa.objectmodel.c.AssumeEdge;
import org.sosy_lab.cpachecker.cfa.objectmodel.c.CallToReturnEdge;
import org.sosy_lab.cpachecker.cfa.objectmodel.c.DeclarationEdge;
import org.sosy_lab.cpachecker.cfa.objectmodel.c.FunctionCallEdge;
import org.sosy_lab.cpachecker.cfa.objectmodel.c.FunctionDefinitionNode;
import org.sosy_lab.cpachecker.cfa.objectmodel.c.FunctionReturnEdge;
import org.sosy_lab.cpachecker.cfa.objectmodel.c.ReturnStatementEdge;
import org.sosy_lab.cpachecker.cfa.objectmodel.c.StatementEdge;
import org.sosy_lab.cpachecker.core.interfaces.AbstractElement;
import org.sosy_lab.cpachecker.core.interfaces.Precision;
import org.sosy_lab.cpachecker.core.interfaces.TransferRelation;
import org.sosy_lab.cpachecker.cpa.pointer.PointerElement;
import org.sosy_lab.cpachecker.exceptions.CPATransferException;
import org.sosy_lab.cpachecker.exceptions.UnrecognizedCCodeException;
import org.sosy_lab.cpachecker.exceptions.UnrecognizedCFAEdgeException;

@Options(prefix="cpa.interval")
public class IntervalAnalysisTransferRelation implements TransferRelation
{

  /**
   * base name of the variable that is introduced to pass results from returning function calls
   */
  private static final String RETURN_VARIABLE_BASE_NAME = "___cpa_temp_result_var_";

  private final Set<String> globalVars = new HashSet<String>();

  @Option
  private int threshold = 0;

  public IntervalAnalysisTransferRelation(Configuration config) throws InvalidConfigurationException
  {
    config.inject(this);
  }

  @Override
  public Collection<? extends AbstractElement> getAbstractSuccessors (AbstractElement element, Precision precision, CFAEdge cfaEdge) throws CPATransferException
  {
    Collection<? extends AbstractElement> successors  = null;

    AbstractElement successor                         = null;

    IntervalAnalysisElement intervalElement           = (IntervalAnalysisElement)element;

    // check the type of the edge
    switch(cfaEdge.getEdgeType())
    {
      // if edge is a statement edge, e.g. a = b + c
      case StatementEdge:
        StatementEdge statementEdge = (StatementEdge)cfaEdge;
        successor = handleStatement(intervalElement, statementEdge.getStatement(), cfaEdge);
        break;

        // this is the statement edge which leads the function to the last node of its CFA (not same as a return edge)
      case ReturnStatementEdge:
        ReturnStatementEdge returnEdge = (ReturnStatementEdge)cfaEdge;
        successor = handleExitFromFunction(intervalElement, returnEdge.getExpression(), returnEdge);
        break;

      // edge is a declaration edge, e.g. int a;
      case DeclarationEdge:
        successor = handleDeclaration(intervalElement, (DeclarationEdge)cfaEdge);
        break;

      // this is an assumption, e.g. if(a == b)
      case AssumeEdge:
        AssumeEdge assumeEdge = (AssumeEdge)cfaEdge;
        successors = handleAssumption(intervalElement, assumeEdge.getExpression(), cfaEdge, assumeEdge.getTruthAssumption());
        break;

      case BlankEdge:
        successor = intervalElement.clone();
        break;

      case FunctionCallEdge:
        FunctionCallEdge functionCallEdge = (FunctionCallEdge)cfaEdge;
        successor = handleFunctionCall(intervalElement, functionCallEdge);
        break;

      // this is a return edge from function, this is different from return statement
      // of the function. See case in statement edge for details
      case FunctionReturnEdge:
        successor = handleFunctionReturn(intervalElement, (FunctionReturnEdge)cfaEdge);
        break;

      default:
        throw new UnrecognizedCFAEdgeException(cfaEdge);
    }

    if(successors != null)
      return successors;

    else if(successor == null)
      return noSuccessors();

    else
      return soleSuccessor(successor);
  }

  /**
   * Handles return from one function to another function.
   *
   * @param element previous abstract element.
   * @param functionReturnEdge return edge from a function to its call site.
   * @return new abstract element.
   */
  private IntervalAnalysisElement handleFunctionReturn(IntervalAnalysisElement element, FunctionReturnEdge functionReturnEdge)
    throws UnrecognizedCCodeException
  {
    CallToReturnEdge summaryEdge = functionReturnEdge.getSuccessor().getEnteringSummaryEdge();

    IASTFunctionCall expression = summaryEdge.getExpression();

    IntervalAnalysisElement newElement = element.getPreviousElement().clone();

    String callerFunctionName = functionReturnEdge.getSuccessor().getFunctionName();
    String calledFunctionName = functionReturnEdge.getPredecessor().getFunctionName();

    // expression is an assignment operation, e.g. a = g(b);
    if(expression instanceof IASTFunctionCallAssignmentStatement)
    {
      IASTFunctionCallAssignmentStatement funcExp = (IASTFunctionCallAssignmentStatement)expression;

      IASTExpression operand1 = funcExp.getLeftHandSide();

      // left hand side of the expression has to be a variable
      if(operand1 instanceof IASTIdExpression)
      {
        String assignedVariableName = operand1.getRawSignature();

        String returnedVariableName = calledFunctionName + "::" + RETURN_VARIABLE_BASE_NAME;

        for(String globalVar : globalVars)
        {
          // if the assigned variable represents global variable, set the global variable to the value of the returning variable or unknown
          if(globalVar.equals(assignedVariableName))
          {
            Interval interval = element.contains(returnedVariableName) ? element.getInterval(returnedVariableName) : Interval.createUnboundInterval();

            newElement.addInterval(globalVar, interval, this.threshold);
          }

          // import the global variables into the scope of the called function
          else
          {
            Interval interval = element.contains(globalVar) ? element.getInterval(globalVar) : Interval.createUnboundInterval();

            newElement.addInterval(globalVar, interval, this.threshold);
          }
        }

        // set the value of the assigned variable to the value of the returned variable
        if(!globalVars.contains(assignedVariableName))
        {
          Interval interval = element.contains(returnedVariableName) ? element.getInterval(returnedVariableName) : Interval.createUnboundInterval();

          newElement.addInterval(constructVariableName(assignedVariableName, callerFunctionName), interval, this.threshold);
        }
      }

      else
        throw new UnrecognizedCCodeException("on function return", summaryEdge, operand1);
    }

    // import the global variables back into the scope of the calling function
    else if(expression instanceof IASTFunctionCallStatement)
    {
      for(String globalVar : globalVars)
      {
          Interval interval = element.contains(globalVar) ? element.getInterval(globalVar) : Interval.createUnboundInterval();

          newElement.addInterval(globalVar, interval, this.threshold);
      }
    }

    else
      throw new UnrecognizedCCodeException("on function return", summaryEdge, expression.asStatement());

    return newElement;
  }

  /**
   * This method handles function calls.
   *
   * @param previousElement the previous element of the analysis, before the function call
   * @param callEdge the respective CFA edge
   * @return the successor element
   * @throws UnrecognizedCCodeException
   */
  private IntervalAnalysisElement handleFunctionCall(IntervalAnalysisElement previousElement, FunctionCallEdge callEdge)
    throws UnrecognizedCCodeException
  {
    FunctionDefinitionNode functionEntryNode = callEdge.getSuccessor();

    String calledFunctionName = functionEntryNode.getFunctionName();
    String callerFunctionName = callEdge.getPredecessor().getFunctionName();

    List<String> parameterNames = functionEntryNode.getFunctionParameterNames();
    List<IASTExpression> arguments  = callEdge.getArguments();

    assert(parameterNames.size() == arguments.size());

    IntervalAnalysisElement newElement = new IntervalAnalysisElement(previousElement);

    // import global variables into the current scope first
    for(String globalVar : globalVars)
    {
      if(previousElement.contains(globalVar))
        newElement.addInterval(globalVar, previousElement.getInterval(globalVar), threshold);
    }

    // set the interval of each formal parameter to the interval of its respective actual parameter
    for(int i = 0; i < arguments.size(); i++)
    {
      Interval interval = evaluateInterval(previousElement, arguments.get(i), callerFunctionName, callEdge);

      String formalParameterName = constructVariableName(parameterNames.get(i), calledFunctionName);

      newElement.addInterval(formalParameterName, interval, this.threshold);
    }

    return newElement;
  }

  /**
   * This method handles the statement edge which leads the function to the last node of its CFA (not same as a return edge).
   *
   * @author loewe
   * @param element the analysis element
   * @param expression the expression
   * @param ReturnStatementEdge the CFA edge corresponding to this statement
   * @return the successor elements
   */
  private IntervalAnalysisElement handleExitFromFunction(IntervalAnalysisElement element, IASTExpression expression, ReturnStatementEdge returnEdge)
    throws UnrecognizedCCodeException
  {
    // assign the value of the function return to a new variable
    return handleAssignmentToVariable(element, RETURN_VARIABLE_BASE_NAME, expression, returnEdge);
  }

  /**
   * This method handles assumptions.
   *
   * @author loewe
   * @param element the analysis element
   * @param expression the expression containing the assumption
   * @param cfaEdge the CFA edge corresponding to this expression
   * @param truthValue flag to determine whether this is the then- or the else-branch of the assumption
   * @return the successor elements
   */
  private Collection<? extends AbstractElement> handleAssumption(IntervalAnalysisElement element, IASTExpression expression, CFAEdge cfaEdge, boolean truthValue)
    throws UnrecognizedCCodeException
  {
    // first, unpack the expression to deal with a raw assumption
    if(expression instanceof IASTUnaryExpression)
    {
      IASTUnaryExpression unaryExp = ((IASTUnaryExpression)expression);

      switch(unaryExp.getOperator())
      {
        // remove negation
        case NOT:
          return handleAssumption(element, unaryExp.getOperand(), cfaEdge, !truthValue);

        default:
          throw new UnrecognizedCCodeException(cfaEdge, unaryExp);
      }
    }

    // a plain (boolean) identifier, e.g. if(a) or [exp1 op exp2]
    else if(expression instanceof IASTIdExpression
        || expression instanceof IASTBinaryExpression)
    {
      // convert if(a) to a BinaryExpression, i.e. if(a != 0)
      if(expression instanceof IASTIdExpression)
      {
        IASTIntegerLiteralExpression zero = new IASTIntegerLiteralExpression("0",
            expression.getFileLocation(),
            expression.getExpressionType(),
            BigInteger.ZERO);

        expression = new IASTBinaryExpression(expression.getRawSignature(),
                                    expression.getFileLocation(),
                                    expression.getExpressionType(),
                                    expression,
                                    zero,
                                    BinaryOperator.NOT_EQUALS);
      }

      // handle the binary expression
      IntervalAnalysisElement newElement = element.clone();

      BinaryOperator operator = ((IASTBinaryExpression)expression).getOperator();
      IASTExpression operand1 = ((IASTBinaryExpression)expression).getOperand1();
      IASTExpression operand2 = ((IASTBinaryExpression)expression).getOperand2();

      String functionName     = cfaEdge.getPredecessor().getFunctionName();

      Interval interval1 = evaluateInterval(newElement, operand1, functionName, cfaEdge);
      Interval interval2 = evaluateInterval(newElement, operand2, functionName, cfaEdge);

      switch(operator)
      {
        case MINUS:
        case PLUS:
          Interval result = null;

          if(operator == BinaryOperator.MINUS)
            result = interval1.minus(interval2);

          else if(operator == BinaryOperator.PLUS)
            result = interval1.plus(interval2);

          // in then-branch and interval maybe true, or in else-branch and interval maybe false, add a successor
          if((truthValue && !result.isFalse()) || (!truthValue && !result.isTrue()))
            return soleSuccessor(newElement);

          else
            return noSuccessors();

        case EQUALS:
        case NOT_EQUALS:
        case GREATER_THAN:
        case GREATER_EQUAL:
        case LESS_THAN:
        case LESS_EQUAL:
          return processAssumption(newElement, operator, operand1, operand2, truthValue, cfaEdge);

        default:
          throw new UnrecognizedCCodeException(cfaEdge, expression);
      }
    }

    return noSuccessors();
  }

  private Collection<? extends AbstractElement> processAssumption(IntervalAnalysisElement element, BinaryOperator operator, IASTExpression operand1, IASTExpression operand2, boolean truthValue, CFAEdge cfaEdge) throws UnrecognizedCCodeException
  {
    if(!truthValue)
      return processAssumption(element, negateOperator(operator), operand1, operand2, !truthValue, cfaEdge);

    Interval orgInterval1 = evaluateInterval(element, operand1, cfaEdge.getPredecessor().getFunctionName(), cfaEdge);
    Interval tmpInterval1 = orgInterval1.clone();

    Interval orgInterval2 = evaluateInterval(element, operand2, cfaEdge.getPredecessor().getFunctionName(), cfaEdge);
    Interval tmpInterval2 = orgInterval2.clone();

    String variableName1 = constructVariableName(operand1.getRawSignature(), cfaEdge.getPredecessor().getFunctionName());
    String variableName2 = constructVariableName(operand2.getRawSignature(), cfaEdge.getPredecessor().getFunctionName());

    // determine whether or not the respective operand is an identifier
    boolean isIdOp1 = operand1 instanceof IASTIdExpression;
    boolean isIdOp2 = operand2 instanceof IASTIdExpression;

    // a < b, a < 1
    if(operator == BinaryOperator.LESS_THAN)
    {
      // a may be less than b, so there can be a successor
      if(tmpInterval1.mayBeLessThan(tmpInterval2))
      {
        if(isIdOp1) element.addInterval(variableName1, orgInterval1.limitUpperBoundBy(tmpInterval2.minus(1)), threshold);
        if(isIdOp2) element.addInterval(variableName2, orgInterval2.limitLowerBoundBy(tmpInterval1.plus(1)), threshold);
      }

      // a is always greater than b, so a can't be less than b, so there can't be a successor
      else
        return noSuccessors();
    }

    // a <= b, a <= 1
    else if(operator == BinaryOperator.LESS_EQUAL)
    {
      // a may be less or equal than b, so there can be a successor
      if(tmpInterval1.mayBeLessOrEqualThan(tmpInterval2))
      {
        if(isIdOp1) element.addInterval(variableName1, orgInterval1.limitUpperBoundBy(tmpInterval2), threshold);
        if(isIdOp2) element.addInterval(variableName2, orgInterval2.limitLowerBoundBy(tmpInterval1), threshold);
      }

      // a is always greater than b, so a can't be less than b, so there can't be a successor
      else
        return noSuccessors();
    }

    // a > b, a > 1
    else if(operator == BinaryOperator.GREATER_THAN)
    {
      // a may be greater than b, so there can be a successor
      if(tmpInterval1.mayBeGreaterThan(tmpInterval2))
      {
        if(isIdOp1) element.addInterval(variableName1, orgInterval1.limitLowerBoundBy(tmpInterval2.plus(1)), threshold);
        if(isIdOp2) element.addInterval(variableName2, orgInterval2.limitUpperBoundBy(tmpInterval1.minus(1)), threshold);
      }

      // a is always less than b, so a can't be greater than b, so there can't be a successor
      else
        return noSuccessors();
    }

    // a >= b, a >= 1
    else if(operator == BinaryOperator.GREATER_EQUAL)
    {
      // a may be greater or equal than b, so there can be a successor
      if(tmpInterval1.mayBeGreaterOrEqualThan(tmpInterval2))
      {
        if(isIdOp1) element.addInterval(variableName1, orgInterval1.limitLowerBoundBy(tmpInterval2), threshold);
        if(isIdOp2) element.addInterval(variableName2, orgInterval2.limitUpperBoundBy(tmpInterval1), threshold);
      }

      // a is always less than b, so a can't be greater than b, so there can't be a successor
      else
        return noSuccessors();
    }

    // a == b, a == 1
    else if(operator == BinaryOperator.EQUALS)
    {
      // a and b intersect, so they may have the same value, so they may be equal
      if(tmpInterval1.intersects(tmpInterval2))
      {
        if(isIdOp1) element.addInterval(variableName1, orgInterval1.intersect(tmpInterval2), threshold);
        if(isIdOp2) element.addInterval(variableName2, orgInterval2.intersect(tmpInterval1), threshold);
      }

      // a and b do not intersect, so they can't be equal, so there can't be a successor
      else
        return noSuccessors();
    }

    // a != b, a != 1
    else if(operator == BinaryOperator.NOT_EQUALS)
    {
      // a = [x, x] = b => a and b are always equal, so there can't be a successor
      if(tmpInterval1.isSingular() && tmpInterval1.equals(tmpInterval2))
        return noSuccessors();

      // TODO: currently depends on the fact that operand1 is a identifier, while operand2 is a literal
      if(isIdOp1 && !isIdOp2)
      {
        IntervalAnalysisElement newElement = null;

        Collection<IntervalAnalysisElement> successors = new LinkedList<IntervalAnalysisElement>();

        Interval result = null;

        if(!(result = orgInterval1.intersect(Interval.createUpperBoundedInterval(orgInterval2.getLow() - 1))).isEmpty())
        {
          newElement = element.clone();

          newElement.addInterval(variableName1, result, threshold);

          successors.add(newElement);
        }

        if(!(result = orgInterval1.intersect(Interval.createLowerBoundedInterval(orgInterval2.getLow() + 1))).isEmpty())
        {
          newElement = element.clone();

          newElement.addInterval(variableName1, result, threshold);

          successors.add(newElement);
        }

        return successors;
      }
    }
    else
      throw new UnrecognizedCCodeException("unknown operator", cfaEdge);

    return soleSuccessor(element);
  }

  /**
   * This method return the negated counter part for a given operator
   *
   * @author loewe
   * @param operator
   * @return the negated counter part of the given operator
   */
  private BinaryOperator negateOperator(BinaryOperator operator)
  {
    switch(operator) {
      case EQUALS:
        return BinaryOperator.NOT_EQUALS;

      case NOT_EQUALS:
        return BinaryOperator.EQUALS;

      case LESS_THAN:
        return BinaryOperator.GREATER_EQUAL;

      case LESS_EQUAL:
        return BinaryOperator.GREATER_THAN;

      case GREATER_EQUAL:
        return BinaryOperator.LESS_THAN;

      case GREATER_THAN:
        return BinaryOperator.LESS_EQUAL;

      default:
        return operator;
    }
  }

  /**
   * This method handles variable declarations.
   *
   * So far, only primitive types are supported, pointers are not supported either.
   *
   * @author loewe
   * @param element the analysis element
   * @param declarationEdge the CFA edge
   * @return the successor element
   */
  private IntervalAnalysisElement handleDeclaration(IntervalAnalysisElement element, DeclarationEdge declarationEdge)
  throws UnrecognizedCCodeException
  {
    IntervalAnalysisElement newElement = element.clone();
    if (declarationEdge.getName() != null) {

        // ignore pointer variables
        if (declarationEdge.getDeclSpecifier() instanceof IASTPointerTypeSpecifier)
          return newElement;

        // if this is a global variable, add it to the list of global variables
        if(declarationEdge.isGlobal())
        {
          globalVars.add(declarationEdge.getName().toString());

          Interval interval;

          IASTInitializer init = declarationEdge.getInitializer();

          // global variables may be initialized explicitly on the spot ...
          if(init instanceof IASTInitializerExpression)
          {
            IASTExpression exp = ((IASTInitializerExpression)init).getExpression();

            interval = evaluateInterval(element, exp, "", declarationEdge);
          }

          // ... or implicitly initialized to 0
          else
            interval = new Interval(0L);

          String varName = constructVariableName(declarationEdge.getName().toString(), "");

          newElement.addInterval(varName, interval, this.threshold);
        }

        // non-global variables are initialized with an unbound interval
        else
        {
          String varName = constructVariableName(declarationEdge.getName().toString(), declarationEdge.getPredecessor().getFunctionName());

          newElement.addInterval(varName, Interval.createUnboundInterval(), this.threshold);
          //newElement.addInterval(varName, new Interval(0L), this.threshold);
        }
    }

    return newElement;
  }

  /**
   * This method handles unary and binary statements.
   *
   * @param element the analysis element
   * @param expression the current expression
   * @param cfaEdge the CFA edge
   * @return the successor
   * @throws UnrecognizedCCodeException
   */
  private IntervalAnalysisElement handleStatement(IntervalAnalysisElement element, IASTStatement expression, CFAEdge cfaEdge)
    throws UnrecognizedCCodeException
  {
    // expression is an assignment operation, e.g. a = b;
    if(expression instanceof IASTAssignment)
      return handleAssignment(element, (IASTAssignment)expression, cfaEdge);

    // ext(); => do nothing
    else if(expression instanceof IASTFunctionCallStatement)
      return element.clone();

    // a; nothing to do
    else if(expression instanceof IASTExpressionStatement)
      return element.clone();

    else
      throw new UnrecognizedCCodeException(cfaEdge, expression);
  }

   /**
   * This method handles assignments.
   *
   * @author loewe
   * @param element the analysis element
   * @param assignExpression the expression containing the binary expression
   * @param declarationEdge the CFA edge
   * @return the successor element
   * TODO pointer dereferencing via strengthening
   */
  private IntervalAnalysisElement handleAssignment(IntervalAnalysisElement element, IASTAssignment assignExpression, CFAEdge cfaEdge)
    throws UnrecognizedCCodeException
  {
    IASTExpression op1 = assignExpression.getLeftHandSide();
    IASTRightHandSide op2 = assignExpression.getRightHandSide();

    // a = ?
    if(op1 instanceof IASTIdExpression)
      return handleAssignmentToVariable(element, op1.getRawSignature(), op2, cfaEdge);

    // TODO: assignment to pointer, *a = ?
    else if(op1 instanceof IASTUnaryExpression && ((IASTUnaryExpression)op1).getOperator() == UnaryOperator.STAR)
      return element.clone();

    // TODO assignment to field, a->b = ?
    else if (op1 instanceof IASTFieldReference)
      return element.clone();

    // TODO assignment to array cell, a[b] = ?
    else if (op1 instanceof IASTArraySubscriptExpression)
      return element.clone();

    else
      throw new UnrecognizedCCodeException("left operand of assignment has to be a variable", cfaEdge, op1);
  }

  /**
   * This method handles the assignment of a variable.
   *
   * @author loewe
   * @param element the analysis element
   * @param lParam the local name of the variable to assign to
   * @param rightExp the assigning expression
   * @param cfaEdge the respective CFA edge
   * @return the successor element
   */
  private IntervalAnalysisElement handleAssignmentToVariable(IntervalAnalysisElement element, String lParam, IASTRightHandSide rightExp, CFAEdge cfaEdge)
    throws UnrecognizedCCodeException
  {
    String functionName = cfaEdge.getPredecessor().getFunctionName();

    // a = 8.2 or "return;" (when rightExp == null)
    if(rightExp instanceof IASTLiteralExpression || rightExp == null)
      return handleAssignmentOfLiteral(element, lParam, (IASTLiteralExpression)rightExp, functionName);

    // a = b
    else if (rightExp instanceof IASTIdExpression)
      return handleAssignmentOfVariable(element, lParam, (IASTIdExpression)rightExp, functionName);

    // a = (cast) ?
    else if(rightExp instanceof IASTCastExpression)
      return handleAssignmentOfCast(element, lParam, (IASTCastExpression)rightExp, cfaEdge);

    // a = -b
    else if(rightExp instanceof IASTUnaryExpression)
      return handleAssignmentOfUnaryExp(element, lParam, (IASTUnaryExpression)rightExp, cfaEdge);

    // a = b op c
    else if(rightExp instanceof IASTBinaryExpression)
    {
      IASTBinaryExpression binExp = (IASTBinaryExpression)rightExp;

      return handleAssignmentOfBinaryExp(element, lParam, binExp.getOperand1(), binExp.getOperand2(), binExp.getOperator(), cfaEdge);
    }

    // TODO: a = func(); or a = b->c; currently, the interval of a is unbound
    else if(rightExp instanceof IASTFunctionCallExpression || rightExp instanceof IASTFieldReference)
      return element.clone().addInterval(constructVariableName(lParam, functionName), Interval.createUnboundInterval(), threshold);

    else
      throw new UnrecognizedCCodeException(cfaEdge, rightExp);
  }

  /**
   * This method handles the assignment of a casted variable to another variable.
   *
   * This method routes back to {@link IntervalAnalysisTransferRelation#handleAssignmentToVariable(IntervalAnalysisElement, String, IASTExpression, CFAEdge)} with the cast operation being removed.
   *
   * @author loewe
   * @param element the analysis element
   * @param lParam the local name of the variable to assign to
   * @param castExp the expression containing the cast
   * @param cfaEdge the respective CFA edge
   * @return the successor element
   */
  private IntervalAnalysisElement handleAssignmentOfCast(IntervalAnalysisElement element, String lParam, IASTCastExpression castExp, CFAEdge cfaEdge)
    throws UnrecognizedCCodeException
  {
    return handleAssignmentToVariable(element, lParam, castExp.getOperand(), cfaEdge);
  }

  /**
   * This method handles the assignment of a unary expressions to a variable.
   *
   * This method routes back to {@link IntervalAnalysisTransferRelation#handleAssignmentToVariable(IntervalAnalysisElement, String, IASTExpression, CFAEdge)} with the cast operation being removed.
   *
   * @author loewe
   * @param element the analysis element
   * @param lParam the local name of the variable to assign to
   * @param unaryExp the expression to evaluate
   * @param cfaEdge the respective CFA edge
   * @return the successor element
   */
  private IntervalAnalysisElement handleAssignmentOfUnaryExp(IntervalAnalysisElement element, String lParam, IASTUnaryExpression unaryExp, CFAEdge cfaEdge)
    throws UnrecognizedCCodeException
  {
    IntervalAnalysisElement newElement = element.clone();

    String functionName = cfaEdge.getPredecessor().getFunctionName();

    // name of the updated variable, so if "a = -b" is handled, lParam is "a"
    String assignedVar = constructVariableName(lParam, functionName);

    UnaryOperator unaryOperator = unaryExp.getOperator();

    // TODO: a = *b
    if(unaryOperator == UnaryOperator.STAR)
      newElement.addInterval(assignedVar, Interval.createUnboundInterval(), this.threshold);

    // a = -b or similar
    else
    {
      Interval interval = evaluateInterval(element, unaryExp, functionName, cfaEdge);

      if(interval == null)
        interval = Interval.createUnboundInterval();

      newElement.addInterval(assignedVar, interval, this.threshold);
    }

    return newElement;
  }

  /**
   * This method handles the assignment of a binary expression to a variable.
   *
   * @author loewe
   * @param element the analysis element
   * @param lParam the local name of the variable to assign to
   * @param lVarInBinaryExp the expression to the left of the operator
   * @param rVarInBinaryExp the expression to the right of the operator
   * @param binaryOperator the binary operator
   * @param cfaEdge the respective CFA edge
   * @return the successor element
   * @throws UnrecognizedCCodeException
   */
  private IntervalAnalysisElement handleAssignmentOfBinaryExp(IntervalAnalysisElement element, String lParam, IASTExpression lVarInBinaryExp, IASTExpression rVarInBinaryExp, BinaryOperator binaryOperator, CFAEdge cfaEdge)
    throws UnrecognizedCCodeException
  {
    String functionName = cfaEdge.getPredecessor().getFunctionName();

    // name of the updated variable, so if "a = b + c" is handled, lParam is "a"
    String assignedVar = constructVariableName(lParam, functionName);

    IntervalAnalysisElement newElement = element.clone();

    switch (binaryOperator)
    {
      case DIVIDE:
      case MODULO:
      case LESS_EQUAL:
      case GREATER_EQUAL:
      case BINARY_AND:
      case BINARY_OR:
        // TODO which cases can be handled?
        newElement.addInterval(assignedVar, Interval.createUnboundInterval(), this.threshold);
        break;

      case PLUS:
      case MINUS:
      case MULTIPLY:

        Interval interval1 = null;
        Interval interval2 = null;

        // a = *b + c
        // TODO prepare for using strengthen operator to dereference pointer
        if(!(lVarInBinaryExp instanceof IASTUnaryExpression) || ((IASTUnaryExpression)lVarInBinaryExp).getOperator() != UnaryOperator.STAR)
          interval1 = evaluateInterval(element, lVarInBinaryExp, functionName, cfaEdge);

        if(interval1 != null)
          interval2 = evaluateInterval(element, rVarInBinaryExp, functionName, cfaEdge);

        if(interval1 == null || interval2 == null)
          newElement.addInterval(assignedVar, Interval.createUnboundInterval(), this.threshold);

        else
        {
          Interval interval;
          switch (binaryOperator)
          {
            case PLUS:
              interval = interval1.plus(interval2);
              break;

            case MINUS:
              interval = interval1.minus(interval2);
              break;

            case MULTIPLY:
              interval = interval1.times(interval2);
              break;

            default:
              throw new UnrecognizedCCodeException("unkown binary operator", cfaEdge);
          }

          newElement.addInterval(assignedVar, interval, this.threshold);
        }
    }

    return newElement;
  }

  /**
   * This method evaluates an expression and returns the respective interval.
   *
   * @author loewe
   * @param element the analysis element
   * @param expression the expression containing the expression to be evaluated
   * @param functionName the name of the function currently being analyzed
   * @param cfaEdge the respective CFA edge
   * @return the interval in respect to the evaluated expression of null, if the expression could not be evaluated properly
   */
  //getExpressionValue
  private Interval evaluateInterval(IntervalAnalysisElement element, IASTExpression expression, String functionName, CFAEdge cfaEdge)
    throws UnrecognizedCCodeException
  {
    if(expression instanceof IASTLiteralExpression)
    {
      Long value = parseLiteral(expression);

      return (value == null) ? Interval.createUnboundInterval() : new Interval(value, value);
    }

    else if(expression instanceof IASTIdExpression)
    {
      String varName = constructVariableName(expression.getRawSignature(), functionName);

      return (element.contains(varName)) ? element.getInterval(varName) : Interval.createUnboundInterval();
    }

    else if(expression instanceof IASTCastExpression)
      return evaluateInterval(element, ((IASTCastExpression)expression).getOperand(), functionName, cfaEdge);

    else if(expression instanceof IASTUnaryExpression)
    {
      IASTUnaryExpression unaryExpression = (IASTUnaryExpression)expression;

      UnaryOperator unaryOperator = unaryExpression.getOperator();
      IASTExpression unaryOperand = unaryExpression.getOperand();

      switch(unaryOperator)
      {
        case MINUS:
          Interval interval = evaluateInterval(element, unaryOperand, functionName, cfaEdge);

          return (interval == null) ? Interval.createUnboundInterval() : interval.negate();

        default:
          throw new UnrecognizedCCodeException("unknown unary operator", cfaEdge, unaryExpression);
      }
    }

    // added for expression "if (! (req_a___0 + 50 == rsp_d___0))" in "systemc/mem_slave_tlm.1.cil.c"
    else if(expression instanceof IASTBinaryExpression)
    {
      IASTBinaryExpression binaryExpression = (IASTBinaryExpression)expression;

      Interval interval1 = evaluateInterval(element, binaryExpression.getOperand1(), functionName, cfaEdge);
      Interval interval2 = evaluateInterval(element, binaryExpression.getOperand2(), functionName, cfaEdge);

      switch(binaryExpression.getOperator())
      {
        case PLUS:
          return interval1.plus(interval2);

        default:
          throw new UnrecognizedCCodeException("unknown binary operator", cfaEdge, binaryExpression);
      }
    }

    // TODO fields, arrays
    else
      throw new UnrecognizedCCodeException(cfaEdge, expression);
  }

  /**
   * This method handles the assignment of a variable to another variable.
   *
   * If a interval is tracked for the assigning variable, then the assigned variable will be associated with this interval, otherwise, a previously tracked interval for the assigned variable will be deleted.
   *
   * @author loewe
   * @param element the analysis element
   * @param lParam the local name of the variable to assign to
   * @param op2 the expression representing the assigning variable
   * @param functionName the name of the function currently being analyzed
   * @return the successor element
   */
  private IntervalAnalysisElement handleAssignmentOfVariable(IntervalAnalysisElement element, String lParam, IASTExpression op2, String functionName)
  {
    IntervalAnalysisElement newElement = element.clone();

    String assignedVar  = constructVariableName(lParam, functionName);
    String assigningVar = constructVariableName(op2.getRawSignature(), functionName);

    // the new interval is either the one of the assigning variable, or unbound if the assigning variable has to interval associated
    Interval interval = (newElement.contains(assigningVar)) ? newElement.getInterval(assigningVar) : Interval.createUnboundInterval();

    newElement.addInterval(assignedVar, interval, this.threshold);

    return newElement;
  }

  /**
   * This method handles the assignment of a literal to a variable.
   *
   * @author loewe
   * @param element the analysis element
   * @param lParam the local name of the variable to assign to
   * @param op2 the expression representing the literal
   * @param functionName the name of the function currently being analyzed
   * @return the successor element
   */
  private IntervalAnalysisElement handleAssignmentOfLiteral(IntervalAnalysisElement element, String lParam, IASTExpression op2, String functionName)
    throws UnrecognizedCCodeException
  {
    IntervalAnalysisElement newElement = element.clone();

    String variableName = constructVariableName(lParam, functionName);

    // op2 may be null if this is a "return;" statement
    Long value = (op2 == null) ? 0L : parseLiteral(op2);

    // the interval is either unbound or represents an explicit value
    Interval interval = (value == null) ? Interval.createUnboundInterval() : new Interval(value);

    newElement.addInterval(variableName, interval, threshold);

    return newElement;
  }

  /**
   * This method parses an expression to retrieve its literal value.
   *
   * @param expression the expression to parse
   * @return a number or null if the parsing failed
   * @throws UnrecognizedCCodeException
   */
  private Long parseLiteral(IASTExpression expression) throws UnrecognizedCCodeException
  {
    if(expression instanceof IASTLiteralExpression)
    {
      int typeOfLiteral = ((IASTLiteralExpression)expression).getKind();

      if(typeOfLiteral == IASTLiteralExpression.lk_integer_constant)
      {
        String s = expression.getRawSignature();
        if(s.endsWith("L") || s.endsWith("U") || s.endsWith("UL"))
        {
          s = s.replace("L", "");
          s = s.replace("U", "");
          s = s.replace("UL", "");
        }
        try
        {
          return Long.valueOf(s);
        }
        catch (NumberFormatException e)
        {
          throw new UnrecognizedCCodeException("invalid integer literal", null, expression);
        }
      }

      if(typeOfLiteral == IASTLiteralExpression.lk_string_literal)
        return (long)expression.hashCode();
    }

    return null;
  }

  /**
   * This method created a scoped variable name.
   *
   * @param variableName
   * @param functionName
   * @return a scoped variable name
   */
  public String constructVariableName(String variableName, String functionName)
  {
    if(globalVars.contains(variableName))
      return variableName;

    return functionName + "::" + variableName;
  }

  @Override
  public Collection<? extends AbstractElement> strengthen(AbstractElement element, List<AbstractElement> elements,
      CFAEdge cfaEdge, Precision precision) throws UnrecognizedCCodeException
  {
    assert element instanceof IntervalAnalysisElement;
    IntervalAnalysisElement intervalElement = (IntervalAnalysisElement)element;

    for(AbstractElement elem : elements)
    {
      if(elem instanceof PointerElement)
        return strengthen(intervalElement, (PointerElement)elem, cfaEdge, precision);
    }

    return null;
  }

  private Collection<? extends AbstractElement> strengthen(IntervalAnalysisElement intervalElement,
      PointerElement pointerElement, CFAEdge cfaEdge, Precision precision) throws UnrecognizedCCodeException
  {
    return null;
  }

  private Collection<? extends AbstractElement> soleSuccessor(AbstractElement successor)
  {
    return Collections.singleton(successor);
  }

  private Collection<? extends AbstractElement> noSuccessors()
  {
    return Collections.emptySet();
  }
}

